/* -*-c-*-
 *
 * $Id: mptext.c,v 1.1 1999/11/17 18:02:16 mdw Exp $
 *
 * Textual representation of multiprecision numbers
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mptext.c,v $
 * Revision 1.1  1999/11/17 18:02:16  mdw
 * New multiprecision integer arithmetic suite.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <ctype.h>
#include <stdio.h>

#include <mLib/darray.h>

#include "mp.h"
#include "mptext.h"

/*----- Data structures ---------------------------------------------------*/

#ifndef CHARV
   DA_DECL(charv, char);
#  define CHARV
#endif

/*----- Main code ---------------------------------------------------------*/

/* --- @mp_read@ --- *
 *
 * Arguments:	@mp *m@ = destination multiprecision number
 *		@int radix@ = base to assume for data (or zero to guess)
 *		@const mptext_ops *ops@ = pointer to operations block
 *		@void *p@ = data for the operations block
 *
 * Returns:	The integer read, or zero if it didn't work.
 *
 * Use:		Reads an integer from some source.  If the @radix@ is
 *		specified, the number is assumed to be given in that radix,
 *		with the letters `a' (either upper- or lower-case) upwards
 *		standing for digits greater than 9.  Otherwise, base 10 is
 *		assumed unless the number starts with `0' (octal), `0x' (hex)
 *		or `nnn_' (base `nnn').  An arbitrary amount of whitespace
 *		before the number is ignored.
 */

mp *mp_read(mp *m, int radix, const mptext_ops *ops, void *p)
{
  int r;
  int ch;
  unsigned f = 0;

  enum {
    f_neg = 1u,
    f_ok = 2u
  };

  /* --- Initialize the destination number --- */

  MP_MODIFY(m, 16);
  m->vl = m->v;
  m->f &= ~MP_UNDEF;

  /* --- Read an initial character --- */

  ch = ops->get(p);
  while (isspace(ch))
    ch = ops->get(p);

  /* --- Handle an initial sign --- */

  if (ch == '-') {
    f |= f_neg;
    ch = ops->get(p);
    while (isspace(ch))
      ch = ops->get(p);
  }

  /* --- If the radix is zero, look for leading zeros --- */

  if (radix)
    r = -1;
  else if (ch != '0') {
    radix = 10;
    r = 0;
  } else {
    ch = ops->get(p);
    if (ch == 'x') {
      ch = ops->get(p);
      radix = 16;
    } else {
      radix = 8;
      f |= f_ok;
    }
    r = -1;
  }

  /* --- Time to start --- */

  for (;; ch = ops->get(p)) {
    int x;

    /* --- An underscore indicates a numbered base --- */

    if (ch == '_' && r > 0 && r <= 36) {
      radix = r;
      m->vl = m->v;
      r = -1;
      f &= ~f_ok;
      continue;
    }

    /* --- Check that the character is a digit and in range --- */

    if (!isalnum(ch))
      break;
    if (ch >= '0' && ch <= '9')
      x = ch - '0';
    else {
      ch = tolower(ch);
      if (ch >= 'a' && ch <= 'z')	/* ASCII dependent! */
	x = ch - 'a' + 10;
      else
	break;
    }

    /* --- Sort out what to do with the character --- */

    if (x >= 10 && r >= 0)
      r = -1;
    if (x >= radix)
      break;

    if (r >= 0)
      r = r * 10 + x;

    /* --- Stick the character on the end of my integer --- */

    MP_ENSURE(m, MP_LEN(m) + 1);
    MPX_UMULN(m->v, m->vl, m->v, m->vl - 1, radix);
    MPX_UADDN(m->v, m->vl, x);
    MP_SHRINK(m);
    f |= f_ok;
  }

  ops->unget(ch, p);

  /* --- Bail out if the number was bad --- */

  if (!(f & f_ok)) {
    MP_DROP(m);
    return (0);
  }

  /* --- Set the sign and return --- */

  m->f = 0;
  if (f & f_neg)
    m->f |= MP_NEG;
  return (m);
}

/* --- @mp_write@ --- *
 *
 * Arguments:	@mp *m@ = pointer to a multi-precision integer
 *		@int radix@ = radix to use when writing the number out
 *		@const mptext_ops *ops@ = pointer to an operations block
 *		@void *p@ = data for the operations block
 *
 * Returns:	Zero if it worked, nonzero otherwise.
 *
 * Use:		Writes a large integer in textual form.
 */

int mp_write(mp *m, int radix, const mptext_ops *ops, void *p)
{
  charv v = DA_INIT;
  mpw b = radix;
  mp bb;

  /* --- Set various things up --- */

  m = MP_COPY(m);
  mp_build(&bb, &b, &b + 1);

  /* --- If the number is negative, sort that out --- */

  if (m->f & MP_NEG) {
    if (ops->put("-", 1, p))
      return (EOF);
    MP_SPLIT(m);
    m->f &= ~MP_NEG;
  }

  /* --- Write digits to a temporary array --- */

  do {
    mp *q = MP_NEW, *r = MP_NEW;
    int ch;
    mpw x;

    mp_div(&q, &r, m, &bb);
    x = r->v[0];
    if (x < 10)
      ch = '0' + x;
    else
      ch = 'a' + x - 10;
    DA_UNSHIFT(&v, ch);
    MP_DROP(m);
    MP_DROP(r);
    m = q;
  } while (MP_CMP(m, !=, MP_ZERO));

  /* --- Finished that --- */

  MP_DROP(m);

  {
    int rc = ops->put(DA(&v), DA_LEN(&v), p);
    DA_DESTROY(&v);
    return (rc ? EOF : 0);
  }
}

/*----- Test rig ----------------------------------------------------------*/

#ifdef TEST_RIG

#include <mLib/testrig.h>

static int verify(dstr *v)
{
  int ok = 1;
  int ib = *(int *)v[0].buf, ob = *(int *)v[2].buf;
  dstr d = DSTR_INIT;
  mp *m = mp_readdstr(MP_NEW, &v[1], 0, ib);
  if (m) {
    if (!ob) {
      fprintf(stderr, "*** unexpected successful parse\n"
	              "*** input [%i] = %s\n",
	      ib, v[1].buf);
      mp_writedstr(m, &d, 10);
      fprintf(stderr, "*** (value = %s)\n", d.buf);
      ok = 0;
    } else {
      mp_writedstr(m, &d, ob);
      if (d.len != v[3].len || memcmp(d.buf, v[3].buf, d.len) != 0) {
	fprintf(stderr, "*** failed read or write\n"
		        "*** input [%i]    = %s\n"
		        "*** output [%i]   = %s\n"
		        "*** expected [%i] = %s\n",
		ib, v[1].buf, ob, d.buf, ob, v[3].buf);
	ok = 0;
      }
    }
    mp_drop(m);
  } else {
    if (ob) {
      fprintf(stderr, "*** unexpected parse failure\n"
	              "*** input [%i] = %s\n"
	              "*** expected [%i] = %s\n",
	      ib, v[1].buf, ob, v[3].buf);
      ok = 0;
    }
  }

  dstr_destroy(&d);
  return (ok);
}

static test_chunk tests[] = {
  { "mptext", verify,
    { &type_int, &type_string, &type_int, &type_string, 0 } },
  { 0, 0, { 0 } }
};

int main(int argc, char *argv[])
{
  sub_init();
  test_run(argc, argv, tests, SRCDIR "/tests/mptext");
  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
